#include "newPmacV3Axis.h"
#include "asynOctetSyncIO.h"
#include "newPmacV3Controller.h"
#include <cmath>
#include <errlog.h>
#include <limits>
#include <math.h>
#include <string.h>
#include <unistd.h>

newPmacV3Axis::newPmacV3Axis(newPmacV3Controller *pC, int axisNo)
    : asynMotorAxis(pC, axisNo), pC_(pC) {

    static const char *functionName = "newPmacV3Axis::newPmacV3Axis";
    asynStatus status = asynSuccess;

    /*
    The superclass constructor SINQAxis calls in turn its superclass constructor
    asynMotorAxis. In the latter, a pointer to the constructed object this is
    stored inside the array pAxes_:

        pC->pAxes_[axisNo] = this;

    Therefore, the axes are managed by the controller pC. See C804Controller.cpp
    for further explanation. If axisNo is out of bounds, asynMotorAxis prints an
    error (see
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorAxis.cpp,
    line 40). However, we want the IOC creation to stop completely, since this
    is a configuration error.
    */
    if (axisNo >= pC->numAxes_) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: Axis index %d must be smaller than the "
                  "total number of axes %d. Terminating IOC.",
                  functionName, axisNo_, pC->numAxes_);
        exit(-1);
    }

    // Initialize all member variables
    initial_poll_ = true;
    waitForHandshake_ = false;

    // Wait 10 seconds for the handshake until declaring a timeout
    handshakeTimeout_ = 10;

    // Placeholder, is overwritten later
    time_at_init_poll_ = 0;

    // After 3 idle polls, the parameter library definitely had enough time to
    // be initialized
    timeout_param_lib_init_ = 3 * pC->idlePollPeriod_;

    // Provide initial values for some parameter library entries
    status = pC_->setIntegerParam(axisNo_, pC_->rereadEncoderPosition_, 0);
    if (status != asynSuccess) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: Setting an initial parameter library value "
                  "for rereadEncoderPosition_ in axis %d. Terminating IOC.",
                  functionName, axisNo_);
        exit(-1);
    }

    status = pC_->setDoubleParam(axisNo_, pC_->motorPosition_, 0.0);
    if (status != asynSuccess) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: Setting an initial parameter library value "
                  "for motorPosition_ in axis %d. Terminating IOC.",
                  functionName, axisNo_);
        exit(-1);
    }

    // This value is updated in the poll. Initially, we assume that the motor is
    // not enabled.
    status = pC_->setIntegerParam(axisNo_, pC_->motorEnabled_, 0);
    if (status != asynSuccess) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: Setting an initial parameter library value "
                  "for motorPosition_ in axis %d. Terminating IOC.",
                  functionName, axisNo_);
        exit(-1);
    }
}

newPmacV3Axis::~newPmacV3Axis(void) {
    // Since the controller memory is managed somewhere else, we don't need to
    // clean up the pointer pC here.
}

/*
Read the configuration from the motor control unit and the parameter library.
This operation is only allowed if the motor is not moving
*/
asynStatus newPmacV3Axis::readConfig() {

    // Local variable declaration
    static const char *functionName = "newPmacV3Axis::readConfig";
    asynStatus status = asynSuccess;
    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    int nvals = 0;
    double highLimit = 0.0;
    double lowLimit = 0.0;
    double motorRecResolution = 0.0;
    double position = 0.0;
    int axStatus = 0;

    // =========================================================================

    // Motor resolution from parameter library
    status = pC_->getDoubleParam(axisNo_, pC_->motorRecResolution_,
                                 &motorRecResolution);
    if (status == asynParamUndefined) {
        return asynParamUndefined;
    } else if (status != asynSuccess) {
        return pC_->paramLibAccessFailed(status, functionName,
                                         "motorRecResolution_");
    }

    // Software limits and current position
    snprintf(command, sizeof(command), "P%2.2d00 Q%2.2d10 Q%2.2d13 Q%2.2d14",
             axisNo_, axisNo_, axisNo_, axisNo_);
    status = pC_->writeRead(axisNo_, command, response, true);
    nvals = sscanf(response, "%d %lf %lf %lf", &axStatus, &position, &highLimit,
                   &lowLimit);
    if (pC_->checkNumExpectedReads(4, nvals, functionName, command, response,
                                   axisNo_) != asynSuccess) {
        return asynError;
    }

    // If the motor is not in idle status, do not read the configuration
    if (axStatus != 0) {
        return asynError;
    }

    // Transform from motor to user coordinates
    position = position * motorRecResolution;
    highLimit = highLimit * motorRecResolution;
    lowLimit = lowLimit * motorRecResolution;

    /*
    The axis limits are set as: ({[]})
    where [] are the positive and negative limits set in EPICS/NICOS, {} are the
    software limits set on the MCU and () are the hardware limit switches. In
    other words, the EPICS/NICOS limits must be stricter than the software
    limits on the MCU which in turn should be stricter than the hardware limit
    switches. For example, if the hardware limit switches are at [-10, 10], the
    software limits could be at [-9, 9] and the EPICS / NICOS limits could be at
    [-8, 8]. Therefore, we cannot use the software limits read from the MCU
    directly, but need to shrink them a bit. In this case, we're shrinking them
    by 0.1 mm or 0.1 degree (depending on the axis type) on both sides.
    */
    highLimit = highLimit - 0.1;
    lowLimit = lowLimit + 0.1;

    // Store these values in the parameter library
    status = pC_->setDoubleParam(axisNo_, pC_->motorPosition_, position);
    if (status != asynSuccess) {
        return pC_->paramLibAccessFailed(status, functionName,
                                         "motorPosition_");
    }
    status = pC_->setDoubleParam(axisNo_, pC_->motorLowLimit_, lowLimit);
    if (status != asynSuccess) {
        return pC_->paramLibAccessFailed(status, functionName,
                                         "motorLowLimit_");
    }
    status = pC_->setDoubleParam(axisNo_, pC_->motorHighLimit_, highLimit);
    if (status != asynSuccess) {
        return pC_->paramLibAccessFailed(status, functionName,
                                         "motorHighLimit_");
    }

    // Update the parameter library immediately
    status = callParamCallbacks();
    if (status != asynSuccess) {
        // If we can't communicate with the parameter library, it doesn't make
        // sense to try and upstream this to the user -> Just log the error
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Updating the parameter library failed for axis %d\n",
                  functionName, axisNo_);
        return status;
    }

    return this->readEncoderType();
}

asynStatus newPmacV3Axis::poll(bool *moving) {
    // Local variable declaration
    static const char *functionName = "newPmacV3Axis::poll";
    asynStatus pl_status = asynSuccess;
    asynStatus poll_status = asynSuccess;

    // =========================================================================

    // If this poll is the initial poll, check if the parameter library has
    // already been initialized. If not, force EPCIS to repeat the poll until
    // the initialization is complete (or until a timeout is reached). Once the
    // parameter library has been initialized, read configuration data from the
    // motor controller into it.
    if (initial_poll_) {

        if (time_at_init_poll_ == 0) {
            time_at_init_poll_ = time(NULL);
        }

        if (time(NULL) > (time_at_init_poll_ + timeout_param_lib_init_)) {
            asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                      "%s: FATAL ERROR: Could not initialize the parameter "
                      "library until the timeout of %ld seconds after IOC "
                      "startup. Terminating IOC.",
                      functionName, timeout_param_lib_init_);
            exit(-1);
        }

        poll_status = readConfig();
        if (poll_status == asynSuccess) {
            initial_poll_ = false;
        } else if (poll_status == asynParamUndefined) {
            // Wait for 100 ms until trying the entire poll again
            usleep(100000);
            return poll_status;
        } else {
            // Something else went completly wrong => Abort the program
            asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                      "%s: FATAL ERROR: Reading a value from the parameter "
                      "library failed for axis %d (%s). Terminating",
                      functionName, axisNo_,
                      pC_->stringifyAsynStatus(poll_status));
            exit(-1);
        }
    }

    // The poll function is just a wrapper around pollNoUpdate and
    // handles mainly the callParamCallbacks() function. This wrapper is used
    // to make sure callParamCallbacks() is called in case of a premature
    // return.
    poll_status = newPmacV3Axis::pollNoUpdate(moving);

    // If the poll status is ok, reset the error indicators in the parameter
    // library
    if (poll_status == asynSuccess) {
        pl_status = setIntegerParam(pC_->motorStatusProblem_, false);
        if (pl_status != asynSuccess) {
            pC_->paramLibAccessFailed(pl_status, functionName,
                                      "motorStatusProblem_");
        }
        pl_status = setIntegerParam(pC_->motorStatusCommsError_, false);
        if (pl_status != asynSuccess) {
            pC_->paramLibAccessFailed(pl_status, functionName,
                                      "motorStatusCommsError_");
        }
    }

    // According to the function documentation of asynMotorAxis::poll, this
    // function should be called at the end of a poll implementation.
    pl_status = callParamCallbacks();
    if (pl_status != asynSuccess) {
        // If we can't communicate with the parameter library, it doesn't make
        // sense to try and upstream this to the user -> Just log the error
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Updating the parameter library failed for axis %d\n",
                  functionName, axisNo_);
        poll_status = pl_status;
    }

    return poll_status;
}

// Perform the actual poll
asynStatus newPmacV3Axis::pollNoUpdate(bool *moving) {

    // Return value for the poll
    asynStatus poll_status = asynSuccess;

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    static const char *functionName = "newPmacV3Axis::poll";
    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    int nvals = 0;

    int direction = 0;
    int error = 0;
    int axStatus = 0;
    double currentPosition = 0.0;
    double previousPosition = 0.0;
    double motorRecResolution = 0.0;
    int handshakePerformed = 0;

    // =========================================================================

    // Are we currently waiting for a handshake?
    if (waitForHandshake_) {
        snprintf(command, sizeof(command), "P%2.2d23", axisNo_);
        rw_status = pC_->writeRead(axisNo_, command, response, true);
        if (rw_status != asynSuccess) {
            return rw_status;
        }

        nvals = sscanf(response, "%d", &handshakePerformed);
        if (pC_->checkNumExpectedReads(1, nvals, functionName, command,
                                       response, axisNo_) != asynSuccess) {
            return asynError;
        }

        if (handshakePerformed == 1) {
            // Handshake has been performed successfully -> Continue with the
            // poll
            waitForHandshake_ = false;
        } else {
            // Still waiting for the handshake. This is already part of the
            // movement procedure!
            if (time(NULL) < timeAtHandshake_ + handshakeTimeout_) {
                *moving = true;
                return asynSuccess;
            } else {
                // Timed out when waiting for the handshake
                asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                          "%s: Axis %d timed out when waiting for the "
                          "handshake with the MCU.\n",
                          functionName, axisNo_);
                pl_status = setStringParam(
                    pC_->messageText_, "Handshake timed out. This is a bug.");
                if (pl_status != asynSuccess) {
                    return pC_->paramLibAccessFailed(pl_status, functionName,
                                                     "messageText_");
                }
                waitForHandshake_ = false;
                return asynError;
            }
        }
    }

    // Motor resolution from parameter library
    pl_status = pC_->getDoubleParam(axisNo_, pC_->motorRecResolution_,
                                    &motorRecResolution);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorRecResolution_");
    }

    // Read the previous motor position
    pl_status =
        pC_->getDoubleParam(axisNo_, pC_->motorPosition_, &previousPosition);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorPosition_");
    }

    // Check the axis status (Pxx00) and the current motor position (Qxx10)
    snprintf(command, sizeof(command), "P%2.2d00 Q%2.2d10 P%2.2d01", axisNo_,
             axisNo_, axisNo_);
    rw_status = pC_->writeRead(axisNo_, command, response, true);
    if (rw_status != asynSuccess) {
        return rw_status;
    }

    nvals = sscanf(response, "%d %lf %d", &axStatus, &currentPosition, &error);
    if (pC_->checkNumExpectedReads(3, nvals, functionName, command, response,
                                   axisNo_) != asynSuccess) {
        return asynError;
    }

    // Intepret the status
    switch (axStatus) {
    case -6:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Axis %d is stopping\n", functionName, axisNo_);
        *moving = true;
        break;
    case -5:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Axis %d is deactivated\n", functionName, axisNo_);
        *moving = false;

        pl_status = setStringParam(pC_->messageText_, "Deactivated");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        // No further evaluation of the axis status is necessary
        return asynSuccess;
    case -4:
        asynPrint(
            pC_->pasynUserSelf, ASYN_TRACE_ERROR,
            "%s: Emergency stop has been activated. All axes are stopped.\n",
            functionName);
        *moving = false;

        pl_status = setStringParam(pC_->messageText_, "Emergency stop");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        // No further evaluation of the axis status is necessary
        return asynSuccess;
    case -3:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Axis %d is inhibited\n", functionName, axisNo_);
        *moving = false;

        pl_status = setStringParam(pC_->messageText_, "Disabled");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        // No further evaluation of the axis status is necessary
        return asynSuccess;
    case 0:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Axis %d is ready for movement\n", functionName, axisNo_);
        *moving = false;
        break;
    case 1:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Move order for %d acknowledged\n", functionName,
                  axisNo_);
        *moving = true;
        break;
    case 2:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Move order for %d is possible\n", functionName, axisNo_);
        *moving = true;
        break;
    case 3:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: %d in Air Cushion Outout status\n", functionName,
                  axisNo_);
        *moving = true;
        break;
    case 4:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: %d in Air Cushion Input status\n", functionName,
                  axisNo_);
        *moving = true;
        break;
    case 5:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
                  "%s: Axis %d is moving\n", functionName, axisNo_);
        *moving = true;
        break;
    default:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Reached unreachable state P%2.2d00 = %d.\n",
                  functionName, axisNo_, axStatus);
        pl_status =
            setStringParam(pC_->messageText_,
                           "Unreachable state has been reached. This is a bug");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }
        *moving = false;
    }

    if (*moving) {
        // If the axis is moving, evaluate the movement direction
        if ((currentPosition - previousPosition) > 0) {
            direction = 1;
        } else {
            direction = 0;
        }
    }

    // Error handling
    switch (error) {
    case 0:
        // No error
        break;
    case 1:
        // EPICS should already prevent this issue in the first place,
        // since it contains the user limits
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Target position would exceed user limits in axis "
                  "%d. EPICS should prevent an out-of-bounds target "
                  "position, this is a bug.\n",
                  functionName, axisNo_);

        pl_status =
            setStringParam(pC_->messageText_,
                           "Target position would exceed software limits. This "
                           "is a bug.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    case 5:
        // Command not possible
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Axis %d is still moving, but received another move "
                  "command. EPICS should prevent this, this is a bug.\n",
                  functionName, axisNo_);

        pl_status = setStringParam(pC_->messageText_,
                                   "Axis received move command while it is "
                                   "still moving. This is a bug.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    case 10:
        /*
        Software limits of the controller have been hit. Since the EPICS limits
        are derived from the software limits and are a little bit smaller, this
        error case can only happen if either the axis has an incremental encoder
        which is not properly homed or if a bug occured.
        */
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_WARNING,
                  "%s: Axis %d hit the controller limits. Try homing the axis, "
                  "if that doesn't remove the error, this is a bug.\n",
                  functionName, axisNo_);

        snprintf(command, sizeof(command),
                 "Software limits hit (P%2.2d01 = %d). Try homing the motor. "
                 "If that doesn't work, contact the support",
                 axisNo_, error);
        pl_status = setStringParam(pC_->messageText_, command);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    case 11:
        // Following error
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Maximum allowed following error exceeded for axis %d.\n",
                  functionName, axisNo_);

        snprintf(command, sizeof(command),
                 "Maximum allowed following error exceeded (P%2.2d01 = 11). "
                 "Please contact Electronics support.",
                 axisNo_);
        pl_status = setStringParam(pC_->messageText_, command);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    case 13:
        // Watchdog of the controller final stage triggered
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Driver hardware error triggered for axis %d.\n",
                  functionName, axisNo_);

        snprintf(command, sizeof(command),
                 "Driver hardware error (P%2.2d01 = 13). "
                 "Please contact Electronics support.",
                 axisNo_);
        pl_status = setStringParam(pC_->messageText_, command);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    case 14:
        // EPICS should already prevent this issue in the first place,
        // since it contains the user limits
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Target position would exceed hardware limits in axis "
                  "%d. Homing might be necessary.\n",
                  functionName, axisNo_);

        snprintf(command, sizeof(command),
                 "Move command exceeds hardware limits (P%2.2d01 = %d). Try "
                 "homing the motor. If that doesn't work, contact the support",
                 axisNo_, error);
        pl_status = setStringParam(pC_->messageText_, command);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    default:
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Axis %d reached an unreachable state (P%2.2d01 = %d).\n",
                  functionName, axisNo_, axisNo_, error);

        pl_status = setStringParam(pC_->messageText_,
                                   "Axis reached an unreachable state. Please "
                                   "contact electronics support");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }

        poll_status = asynError;
        break;
    }

    // Update the parameter library
    if (error != 0) {
        pl_status = setIntegerParam(pC_->motorStatusProblem_, true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "motorStatusProblem_");
        }
    }

    if (*moving == false) {
        pl_status = setIntegerParam(pC_->motorMoveToHome_, 0);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "motorMoveToHome_");
        }
    }

    pl_status =
        setIntegerParam(pC_->motorEnabled_, (axStatus != -3 && axStatus != -5));
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorEnabled_");
    }

    pl_status = setIntegerParam(pC_->motorStatusMoving_, *moving);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorStatusMoving_");
    }

    pl_status = setIntegerParam(pC_->motorStatusDone_, !(*moving));
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorStatusDone_");
    }

    pl_status = setIntegerParam(pC_->motorStatusDirection_, direction);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorStatusDirection_");
    }

    pl_status = setDoubleParam(pC_->motorPosition_, currentPosition);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorPosition_");
    }

    return poll_status;
}

asynStatus newPmacV3Axis::move(double position, int relative,
                               double minVelocity, double maxVelocity,
                               double acceleration) {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    static const char *functionName = "newPmacV3Axis::move";
    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    double motorCoordinatesPosition = 0.0;
    int enabled = 0;
    double motorRecResolution = 0.0;

    // =========================================================================

    pl_status = pC_->getIntegerParam(axisNo_, pC_->motorEnabled_, &enabled);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorEnabled_");
    }

    pl_status = pC_->getDoubleParam(axisNo_, pC_->motorRecResolution_,
                                    &motorRecResolution);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "motorRecResolution_");
    }

    if (enabled == 0) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Axis %d is disabled\n", functionName, axisNo_);
        return asynSuccess;
    }

    // Convert from user to motor units
    motorCoordinatesPosition = position / motorRecResolution;

    asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
              "%s: Start of axis %d to position %lf\n", functionName, axisNo_,
              position);

    // Perform handshake, Set target position and start the move command

    if (relative) {
        snprintf(command, sizeof(command), "P%2.2d23=0 Q%2.2d02=%lf M%2.2d=2",
                 axisNo_, axisNo_, motorCoordinatesPosition, axisNo_);
    } else {
        snprintf(command, sizeof(command), "P%2.2d23=0 Q%2.2d01=%lf M%2.2d=1",
                 axisNo_, axisNo_, motorCoordinatesPosition, axisNo_);
    }

    // We don't expect an answer
    rw_status = pC_->writeRead(axisNo_, command, response, false);
    if (rw_status != asynSuccess) {
        asynPrint(
            pC_->pasynUserSelf, ASYN_TRACE_ERROR,
            "%s: Starting movement to target position %lf failed for axis %d\n",
            functionName, position, axisNo_);
        pl_status = setIntegerParam(pC_->motorStatusProblem_, true);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "motorStatusProblem_");
        }
        return rw_status;
    }

    // In the next poll, we will check if the handshake has been performed in a
    // reasonable time
    waitForHandshake_ = true;
    timeAtHandshake_ = time(NULL);

    return rw_status;
}

asynStatus newPmacV3Axis::stop(double acceleration) {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    static const char *functionName = "newPmacV3Axis::stopAxis";
    bool moving = false;
    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];

    // =========================================================================

    pl_status = pollNoUpdate(&moving);
    if (pl_status != asynSuccess) {
        return pl_status;
    }

    if (moving) {
        // only send a stop when actually moving
        snprintf(command, sizeof(command), "M%2.2d=8", axisNo_);
        rw_status = pC_->writeRead(axisNo_, command, response, false);

        if (rw_status != asynSuccess) {
            asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                      "%s: Stopping the movement failed for axis %d\n",
                      functionName, axisNo_);

            pl_status = setIntegerParam(pC_->motorStatusProblem_, true);
            if (pl_status != asynSuccess) {
                return pC_->paramLibAccessFailed(pl_status, functionName,
                                                 "motorStatusProblem_");
            }
        }
    }
    return rw_status;
}

/*
Home the axis. On absolute encoder systems, this is a no-op
*/
asynStatus newPmacV3Axis::home(double min_velocity, double max_velocity,
                               double acceleration, int forwards) {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    static const char *functionName = "newPmacV3Axis::home";

    // =========================================================================

    pl_status =
        pC_->getStringParam(axisNo_, pC_->encoderType_, pC_->MAXBUF_, response);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "encoderType_");
    }

    // Only send the home command if the axis has an incremental encoder
    if (strcmp(response, IncrementalEncoder) == 0) {
        snprintf(command, sizeof(command), "M%2.2d=9", axisNo_);
        rw_status = pC_->writeRead(axisNo_, command, response, false);
        if (rw_status != asynSuccess) {
            return rw_status;
        }

        pl_status = setIntegerParam(pC_->motorMoveToHome_, 1);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "motorMoveToHome_");
        }

        pl_status = setStringParam(pC_->messageText_, "Homing");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, functionName,
                                             "messageText_");
        }
    }

    return rw_status;
}

/*
Read the encoder type and update the parameter library accordingly
*/
asynStatus newPmacV3Axis::readEncoderType() {

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    static const char *functionName = "newPmacV3Axis::readEncoderType";
    int nvals = 0;
    int encoder_id = 0;
    char encoderType[pC_->MAXBUF_];

    // =========================================================================

    // Check if this is an absolute encoder
    snprintf(command, sizeof(command), "I%2.2X04", axisNo_);
    rw_status = pC_->writeRead(axisNo_, command, response, true);
    if (rw_status != asynSuccess) {
        return rw_status;
    }

    int reponse_length = strlen(response);
    if (reponse_length < 3) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: Unexpected reponse '%s' from axis %d on "
                  "controller %s while reading the encoder type. Aborting...\n",
                  functionName, response, axisNo_, pC_->portName);
        return asynError;
    }

    // We are only interested in the last two digits and the last value in
    // the string before the terminator is \r
    nvals = sscanf(response + (reponse_length - 3), "%2X", &encoder_id);
    if (pC_->checkNumExpectedReads(1, nvals, functionName, command, response,
                                   axisNo_) != asynSuccess) {
        return asynError;
    }

    snprintf(command, sizeof(command), "P46");
    rw_status = pC_->writeRead(axisNo_, command, response, true);
    if (rw_status != asynSuccess) {
        return rw_status;
    }
    int number_of_axes = strtol(response, NULL, 10);

    // If true, the encoder is incremental
    if (encoder_id <= number_of_axes) {
        pl_status = setStringParam(pC_->encoderType_, IncrementalEncoder);
    } else {
        pl_status = setStringParam(pC_->encoderType_, AbsoluteEncoder);
    }

    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "encoderType_");
    }
    return asynSuccess;
}

asynStatus newPmacV3Axis::enable(int on) {

    static const char *functionName = "newPmacV3Axis::enable";
    int ax_status = 0;
    int timeout_enable_disable = 2;
    char command[pC_->MAXBUF_], response[pC_->MAXBUF_];
    int nvals = 0;

    // Status of read-write-operations of ASCII commands to the controller
    asynStatus rw_status = asynSuccess;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    // =========================================================================

    // Check if the axis is currently enabled
    snprintf(command, sizeof(command), "P%2.2d00", axisNo_);
    rw_status = pC_->writeRead(axisNo_, command, response, true);
    if (rw_status != asynSuccess) {
        return rw_status;
    }

    nvals = sscanf(response, "%d", &ax_status);
    if (pC_->checkNumExpectedReads(1, nvals, functionName, command, response,
                                   axisNo_) != asynSuccess) {
        return asynError;
    }

    asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
              "%s: Axis status %d, new input %d \n", functionName, ax_status,
              on);

    // Axis is already enabled / disabled and a new enable / disable command
    // was sent => Do nothing
    if ((ax_status != -3) == on) {
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_WARNING,
                  "%s: axis %d on controller %s is already %s\n", functionName,
                  axisNo_, pC_->portName, on ? "enabled" : "disabled");
        return asynSuccess;
    }

    // Enable / disable the axis
    snprintf(command, sizeof(command), "M%2.2d14=%d", axisNo_, on);
    asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
              "%s: %s axis %d on controller %s\n", functionName,
              on ? "Enable" : "Disable", axisNo_, pC_->portName);
    if (on == 0) {
        pl_status = setStringParam(pC_->messageText_, "Disabling ...");
    } else {
        pl_status = setStringParam(pC_->messageText_, "Enabling ...");
    }
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "messageText_");
    }
    rw_status = pC_->writeRead(axisNo_, command, response, false);
    if (rw_status != asynSuccess) {
        return rw_status;
    }

    // Query the axis status every few milliseconds until the axis has been
    // enabled or until the timeout has been reached
    snprintf(command, sizeof(command), "P%2.2d00", axisNo_);
    int startTime = time(NULL);
    while (time(NULL) < startTime + timeout_enable_disable) {

        // Read the axis status
        usleep(100000);
        rw_status = pC_->writeRead(axisNo_, command, response, true);
        if (rw_status != asynSuccess) {
            return rw_status;
        }
        nvals = sscanf(response, "%d", &ax_status);
        if (pC_->checkNumExpectedReads(1, nvals, functionName, command,
                                       response, axisNo_) != asynSuccess) {
            return asynError;
        }

        if ((ax_status != -3) == on) {
            bool moving = false;
            // Perform a poll to update the parameter library
            poll(&moving);
            return asynSuccess;
        }
    }

    // Failed to change axis status within timeout_enable_disable => Send a
    // corresponding message
    asynPrint(pC_->pasynUserSelf, ASYN_TRACE_FLOW,
              "%s: Failed to %s axis %d on controller %s within %d seconds\n",
              functionName, on ? "enable" : "disable", axisNo_, pC_->portName,
              timeout_enable_disable);

    // Output message to user
    snprintf(command, sizeof(command), "Failed to %s within %d seconds",
             on ? "enable" : "disable", timeout_enable_disable);
    pl_status = setStringParam(pC_->messageText_, "Enabling ...");
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, functionName,
                                         "messageText_");
    }
    return asynError;
}
