
#include "newPmacV3Controller.h"
#include "asynMotorController.h"
#include "asynOctetSyncIO.h"
#include "newPmacV3Axis.h"
#include <epicsExport.h>
#include <errlog.h>
#include <iocsh.h>
#include <netinet/in.h>
#include <registryFunction.h>
#include <string.h>
#include <unistd.h>

static const char *driverName = "newPmacV3Controller";

// Static pointers (valid for the entire lifetime of the IOC). The number behind
// the strings gives the integer number of each variant (see also method
// stringifyAsynStatus)
static const char *asynSuccessStringified = "success";           // 0
static const char *asynTimeoutStringified = "timeout";           // 1
static const char *asynOverflowStringified = "overflow";         // 2
static const char *asynErrorStringified = "error";               // 3
static const char *asynDisconnectedStringified = "disconnected"; // 4
static const char *asynDisabledStringified = "disabled";         // 5
static const char *asynParamAlreadyExistsStringified =
    "parameter already exists";                                           // 6
static const char *asynParamNotFoundStringified = "parameter not found";  // 7
static const char *asynParamWrongTypeStringified = "wrong type";          // 8
static const char *asynParamBadIndexStringified = "bad index";            // 9
static const char *asynParamUndefinedStringified = "parameter undefined"; // 10
static const char *asynParamInvalidListStringified = "invalid list";      // 11

const double newPmacV3Controller::TIMEOUT_ = 5.0; // seconds

/*
Constructor arguments
- portName:
- lowLevelPortName
- numAxes
- movingPollPeriod: Time between polls when moving (in seconds)
- idlePollPeriod: Time between polls when not moving (in seconds)
*/
newPmacV3Controller::newPmacV3Controller(const char *portName,
                                         const char *lowLevelPortName,
                                         int numAxes, double movingPollPeriod,
                                         double idlePollPeriod,
                                         const int &extraParams)
    : asynMotorController(
          portName, numAxes, NUM_MOTOR_DRIVER_PARAMS + extraParams,
          0, // No additional interfaces beyond those in base class
          0, // No additional callback interfaces beyond those in base class
          ASYN_CANBLOCK | ASYN_MULTIDEVICE,
          1,    // autoconnect
          0, 0) // Default priority and stack size

{

    // Initialization of local variables
    static const char *functionName =
        "newPmacV3Controller::newPmacV3Controller";
    asynStatus status = asynSuccess;

    // Initialization of all member variables
    lowLevelPortUser_ = nullptr;

    // =========================================================================;

    /*
    We try to connect to the port via the port name provided by the constructor.
    If this fails, the function is terminated via exit
    */
    pasynOctetSyncIO->connect(lowLevelPortName, 0, &lowLevelPortUser_, NULL);
    if (status != asynSuccess || lowLevelPortUser_ == nullptr) {
        errlogPrintf("FATAL ERROR: in %s: cannot connect to MCU controller\n. "
                     "Terminating",
                     functionName);
        exit(-1);
    }

    // =========================================================================
    // Create additional PVs

    // MOTOR_MESSAGE_TEXT corresponds to the PV definition inside
    // sinqn_asyn_motor.db. This text is used to forward status messages to
    // NICOS and in turn to the user
    status = createParam("MOTOR_MESSAGE_TEXT", asynParamOctet, &messageText_);
    if (status != asynSuccess) {
        paramLibAccessFailed(status, functionName, "messageText_");
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("ENABLE_AXIS", asynParamInt32, &enableMotor_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("AXIS_ENABLED", asynParamInt32, &motorEnabled_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("ENCODER_TYPE", asynParamOctet, &encoderType_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("REREAD_ENCODER_POSITION", asynParamInt32,
                         &rereadEncoderPosition_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("REREAD_ENCODER_POSITION_RBV", asynParamInt32,
                         &rereadEncoderPositionRBV_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("READ_CONFIG", asynParamInt32, &readConfig_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    status =
        createParam("MOTOR_POSITION_RBV", asynParamFloat64, &motorPositionRBV_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s: FATAL ERROR: unable to create parameter (%s). "
                  "Terminating IOC.\n",
                  functionName, stringifyAsynStatus(status));
        exit(-1);
    }

    /*
    Define the end-of-string of a message coming from the device to EPICS.
    It is not necessary to append a terminator to outgoing messages, since
    the message length is encoded in the message header in the getSetResponse
    method.
    */
    const char *message_from_device =
        "\006"; // Hex-code for ACK (acknowledge) -> Each message from the MCU
                // is terminated by this value
    status = pasynOctetSyncIO->setInputEos(
        lowLevelPortUser_, message_from_device, strlen(message_from_device));
    if (status != asynSuccess) {
        asynPrint(
            this->pasynUserSelf, ASYN_TRACE_ERROR,
            "%s: FATAL ERROR: Unable to set input EOS (%s). Terminating IOC.\n",
            functionName, stringifyAsynStatus(status));
        pasynOctetSyncIO->disconnect(lowLevelPortUser_);
        exit(-1);
    }

    status = startPoller(movingPollPeriod, idlePollPeriod, 1);
    if (status != asynSuccess) {
        asynPrint(
            lowLevelPortUser_, ASYN_TRACE_ERROR,
            "%s: FATAL ERROR: Could not start poller (%s). Terminating IOC.\n",
            functionName, stringifyAsynStatus(status));
        exit(-1);
    }
    status = callParamCallbacks();
    if (status != asynSuccess) {
        asynPrint(
            lowLevelPortUser_, ASYN_TRACE_ERROR,
            "%s: FATAL ERROR: Could not start poller (%s). Terminating IOC.\n",
            functionName, stringifyAsynStatus(status));
        exit(-1);
    }
}

newPmacV3Controller::~newPmacV3Controller(void) {
    /*
    Cleanup of the memory allocated in the asynMotorController constructor
    */
    free(this->pAxes_);
}

/*
Access one of the axes of the controller via the axis adress stored in asynUser.
If the axis does not exist or is not a Axis, a nullptr is returned and an
error is emitted.
*/
newPmacV3Axis *newPmacV3Controller::getAxis(asynUser *pasynUser) {
    asynMotorAxis *asynAxis = asynMotorController::getAxis(pasynUser);
    return newPmacV3Controller::castToAxis(asynAxis);
}

/*
Access one of the axes of the controller via the axis index.
If the axis does not exist or is not a Axis, the function must return Null
*/
newPmacV3Axis *newPmacV3Controller::getAxis(int axisNo) {
    asynMotorAxis *asynAxis = asynMotorController::getAxis(axisNo);
    return newPmacV3Controller::castToAxis(asynAxis);
}

newPmacV3Axis *newPmacV3Controller::castToAxis(asynMotorAxis *asynAxis) {
    static const char *functionName = "newPmacV3Controller::getAxis";

    // =========================================================================

    // If the axis slot of the pAxes_ array is empty, a nullptr must be returned
    if (asynAxis == nullptr) {
        return nullptr;
    }

    // Here, an error is emitted since asyn_axis is not a nullptr but also not
    // an instance of Axis
    newPmacV3Axis *axis = dynamic_cast<newPmacV3Axis *>(asynAxis);
    if (axis == nullptr) {
        asynPrint(lowLevelPortUser_, ASYN_TRACE_ERROR,
                  "%s: Axis %d is not an instance of newPmacV3Axis",
                  functionName, axis->axisNo_);
    }
    return axis;
}

/*
Sends the given command to the axis specified by axisNo and returns the response
of the axis.
*/
asynStatus newPmacV3Controller::writeRead(int axisNo, const char *command,
                                          char *response,
                                          bool expect_response) {
    // Definition of local variables.
    static const char *functionName = "newPmacV3Controller::writeRead";
    asynStatus status = asynSuccess;
    asynStatus pl_status = asynSuccess;
    char full_command[MAXBUF_] = {0};
    char user_message[MAXBUF_] = {0};
    int motorStatusProblem = 0;

    // Send the message and block the thread until either a response has been
    // received or the timeout is triggered
    int eomReason = 0; // Flag indicating why the message has ended
    // Number of bytes of the outgoing message (which is command + the
    // end-of-string terminator defined in the constructor)
    size_t nbytesOut = 0;
    // Number of bytes of the incoming message (which is response + the
    // end-of-string terminator defined in the constructor)
    size_t nbytesIn = 0;

    // =========================================================================

    newPmacV3Axis *axis = getAxis(axisNo);
    if (axis == nullptr) {
        // We already did the error logging directly in getAxis
        return asynError;
    }

    /*
    The message protocol of the pmacV3 used at PSI looks as follows (all
    characters immediately following each other without a newline):
    0x40 (ASCII value of @) -> Request for download
    0xBF (ASCII value of ¿) -> Select mode "get_response"
    0x00 (ASCII value of 0)
    0x00 (ASCII value of 0)
    0x00 (ASCII value of 0)
    0x00 (ASCII value of 0)
    0x00 (ASCII value of 0)
    [message length in network byte order] -> Use the htons function for this
    value [Actual message] It is not necessary to append a terminator, since
    this protocol encodes the message length at the beginning. See Turbo PMAC
    User Manual, page 418 in VR_PMAC_GETRESPONSE

    The message has to be build manually into the buffer full_command, since it
    contains NULL terminators in its middle, therefore the string manipulation
    methods of C don't work.
    */

    // The entire message is equal to the command length
    const size_t commandLength =
        strlen(command) + 1; // +1 because of the appended /r
    const int offset = 8;

    // Positions 2 to 6 must have the value 0. Since full_command is initialized
    // as an array of zeros, we don't need to set these bits manually.
    full_command[0] = '\x40';
    full_command[1] = '\xBF';
    full_command[7] = commandLength;

    snprintf((char *)full_command + offset, MAXBUF_ - offset, "%s\r", command);
    asynPrint(lowLevelPortUser_, ASYN_TRACEIO_DRIVER,
              "%s: Sending command: %s\n", functionName, full_command);

    // Perform the actual writeRead
    status = pasynOctetSyncIO->writeRead(
        lowLevelPortUser_, full_command, commandLength + offset, response,
        MAXBUF_, TIMEOUT_, &nbytesOut, &nbytesIn, &eomReason);

    /*
    If we expect a response, check if we got one. If no response was received,
    flush the PMAC and try again. If that fails as well, return an error
    */
    if (expect_response && strlen(response) == 0) {
        // Flush message as defined in Turbo PMAC User Manual, p. 430:
        // \x40\xB3000
        // VR_DOWNLOAD = \x40
        // VR_PMAC_FLUSH = \xB3
        char flush_msg[5] = {0};
        flush_msg[0] = '\x40';
        flush_msg[1] = '\xB3';
        size_t nbytesOut = 0;
        status = pasynOctetSyncIO->write(lowLevelPortUser_, flush_msg, 5,
                                         TIMEOUT_, &nbytesOut);

        // Wait after the flush so the MCU has time to prepare for the
        // next command
        usleep(100000);

        if (status == asynSuccess) {
            // If flushing the MCU succeded, try to send the command again
            status = pasynOctetSyncIO->writeRead(
                lowLevelPortUser_, full_command, commandLength + offset,
                response, MAXBUF_, TIMEOUT_, &nbytesOut, &nbytesIn, &eomReason);

            // If the command returned an empty string for the second time, give
            // up and propagate the error.
            if (strlen(response) == 0) {
                status = asynError;
            }
        } else {
            asynPrint(pasynUserSelf, ASYN_TRACE_ERROR,
                      "%s: Unable to flush MCU (%s).\n", functionName,
                      stringifyAsynStatus(status));
        }
    }

    // Create custom error messages for different failure modes
    switch (status) {
    case asynSuccess:
        break; // Communicate nothing
    case asynTimeout:
        snprintf(user_message, sizeof(user_message),
                 "connection timeout for axis %d", axisNo);
        break;
    case asynDisconnected:
        snprintf(user_message, sizeof(user_message), "axis is not connected");
        break;
    case asynDisabled:
        snprintf(user_message, sizeof(user_message), "axis is disabled");
        break;
    default:
        snprintf(user_message, sizeof(user_message),
                 "Communication failed (%s)", stringifyAsynStatus(status));
        break;
    }

    if (status != asynSuccess) {
        // Check if the axis already is in an error communication mode. If it is
        // not, upstream the error. This is done to avoid "flooding" the user
        // with different error messages if more than one error ocurred before
        // an error-free communication

        pl_status =
            getIntegerParam(axisNo, motorStatusProblem_, &motorStatusProblem);
        if (pl_status != asynSuccess) {
            return paramLibAccessFailed(pl_status, functionName,
                                        "motorStatusProblem_");
        }

        if (motorStatusProblem == 0) {
            pl_status = axis->setStringParam(this->messageText_, user_message);
            if (pl_status != asynSuccess) {
                return paramLibAccessFailed(pl_status, functionName,
                                            "messageText_");
            }
        }
    }

    // Log the overall status (communication successfull or not)
    if (status == asynSuccess) {
        asynPrint(lowLevelPortUser_, ASYN_TRACEIO_DRIVER,
                  "%s: device response: %s\n", functionName, response);
        pl_status = axis->setIntegerParam(this->motorStatusCommsError_, 0);
    } else {
        asynPrint(lowLevelPortUser_, ASYN_TRACE_ERROR,
                  "%s: asynOctetSyncIO->writeRead failed for command %s on "
                  "axis %d (%s)\n",
                  functionName, command, axisNo, stringifyAsynStatus(status));
        pl_status = axis->setIntegerParam(this->motorStatusCommsError_, 1);
    }

    if (pl_status != asynSuccess) {
        return paramLibAccessFailed(pl_status, functionName,
                                    "motorStatusCommsError_");
    }

    return asynSuccess;
}

asynStatus newPmacV3Controller::writeInt32(asynUser *pasynUser,
                                           epicsInt32 value) {

    int function = pasynUser->reason;
    asynStatus status = asynSuccess;
    static const char *functionName = "newPmacV3Controller::writeInt32";

    // =========================================================================

    newPmacV3Axis *axis = getAxis(pasynUser);
    if (axis == nullptr) {
        // We already did the error logging directly in getAxis
        return asynError;
    }

    // Handle custom PVs
    if (function == enableMotor_) {
        return axis->enable(value);

    } else if (function == rereadEncoderPosition_) {

        char encoderType[MAXBUF_] = {0};

        /*
        This is not a command that can always be run when enabling a
        motor as it also causes relative encoders to reread a position
        necessitating recalibration. We only want it to run on absolute
        encoders. We also want it to be clear to instrument scientists, that
        power has to be cut to the motor, in order to reread the encoder as not
        all motors have breaks and they may start to move when disabled. For
        that reason, we don't automatically disable the motors to run the
        command and instead require that the scientists first disable the motor.
        */

        if (!value) {
            return asynSuccess;
        }

        // Poll the current status of the axis
        bool moving = false;
        status = axis->poll(&moving);
        if (status != asynSuccess) {
            return status;
        }

        // Check if this is an absolute encoder
        status = axis->readEncoderType();
        if (status != asynSuccess) {
            return status;
        }

        status = getStringParam(axis->axisNo_, encoderType_, encoderType);
        if (status != asynSuccess) {
            return paramLibAccessFailed(status, functionName, "encoderType_");
        }

        // Abort if the axis is incremental
        if (strcmp(encoderType, IncrementalEncoder) == 0) {
            asynPrint(this->pasynUserSelf, ASYN_TRACE_WARNING,
                      "%s: Trying to reread absolute encoder of axis %d on "
                      "controller %s, but it is a relative encoder.\n",
                      functionName, axis->axisNo_, portName);
            status = setStringParam(messageText_,
                                    "Cannot reread an incremental encoder.");
            if (status != asynSuccess) {
                return paramLibAccessFailed(status, functionName,
                                            "messageText_");
            }
            return asynError;
        }

        // Check if the axis is disabled. If not, inform the user that this
        // is necessary
        int enabled = 0;
        status = getIntegerParam(axis->axisNo_, motorEnabled_, &enabled);
        if (status != asynSuccess) {
            return paramLibAccessFailed(status, functionName, "motorEnabled_");
        }

        if (enabled == 1) {
            asynPrint(this->pasynUserSelf, ASYN_TRACE_WARNING,
                      "%s: Axis %d on controller %s must be disabled before "
                      "rereading the encoder.\n",
                      functionName, axis->axisNo_, portName);
            status = setStringParam(
                messageText_,
                "Axis must be disabled before rereading the encoder.");
            if (status != asynSuccess) {
                return paramLibAccessFailed(status, functionName,
                                            "messageText_");
            }
            return asynError;
        }

        // Switching on the axis while the rereading process is still ongoing
        // causes it to fail. We currently have no way to check if it is
        // actually finished, so we instead wait for 0.5 seconds.
        usleep(500000);

        // turn off parameter as finished rereading
        // this will only be immediately noticed in the read back variable
        // though
        status = axis->setIntegerParam(rereadEncoderPosition_, 0);
        if (status != asynSuccess) {
            return paramLibAccessFailed(status, functionName,
                                        "rereadEncoderPosition_");
        }
        return asynSuccess;

    } else {

        return asynMotorController::writeInt32(pasynUser, value);
    }
}

/*
Overloaded from asynMotorController because the special cases "motor enabling"
and "rereading the encoder" must be covered.
*/
asynStatus newPmacV3Controller::readInt32(asynUser *pasynUser,
                                          epicsInt32 *value) {

    int function = pasynUser->reason;
    asynStatus status = asynError;
    static const char *functionName = "newPmacV3Controller::readInt32";

    // =====================================================================

    newPmacV3Axis *axis = getAxis(pasynUser);
    if (axis == nullptr) {
        // We already did the error logging directly in getAxis
        return asynError;
    }

    if (function == rereadEncoderPositionRBV_) {
        // Readback value for rereadEncoderPosition
        status = getIntegerParam(axis->axisNo_, rereadEncoderPosition_, value);
        if (status != asynSuccess) {
            return paramLibAccessFailed(status, functionName,
                                        "rereadEncoderPosition_");
        }
        status =
            setIntegerParam(axis->axisNo_, rereadEncoderPositionRBV_, *value);
        if (status != asynSuccess) {
            return paramLibAccessFailed(status, functionName,
                                        "rereadEncoderPositionRBV_");
        }

        // Update the PVs from the parameter library
        return callParamCallbacks();
    } else {
        return asynMotorController::readInt32(pasynUser, value);
    }
}

const char *newPmacV3Controller::stringifyAsynStatus(asynStatus status) {
    // See
    // https://github.com/epics-modules/asyn/blob/master/asyn/asynDriver/asynDriver.h
    // and
    // https://github.com/epics-modules/asyn/blob/master/asyn/asynPortDriver/paramErrors.h
    // for the definition of the error codes
    switch (status) {
    case asynSuccess:
        return asynSuccessStringified;
    case asynTimeout:
        return asynTimeoutStringified;
    case asynOverflow:
        return asynOverflowStringified;
    case asynError:
        return asynErrorStringified;
    case asynDisconnected:
        return asynDisconnectedStringified;
    case asynDisabled:
        return asynDisabledStringified;
    case asynParamAlreadyExists:
        return asynParamAlreadyExistsStringified;
    case asynParamNotFound:
        return asynParamNotFoundStringified;
    case asynParamWrongType:
        return asynParamWrongTypeStringified;
    case asynParamBadIndex:
        return asynParamBadIndexStringified;
    case asynParamUndefined:
        return asynParamUndefinedStringified;
    case asynParamInvalidList:
        return asynParamInvalidListStringified;
    }

    asynPrint(this->lowLevelPortUser_, ASYN_TRACE_ERROR,
              "newPmacV3Controller::stringifyAsynStatus: FATAL error: Reached "
              "unreachable code. Terminating\n");
    exit(-1);
}

asynStatus newPmacV3Controller::paramLibAccessFailed(asynStatus status,
                                                     const char *functionName,
                                                     const char *parameter) {
    char message[MAXBUF_] = {0};
    snprintf(message, sizeof(message),
             "Accessing the parameter library failed for parameter %s (%s). "
             "This is a bug, please inform the software support.\n",
             parameter, stringifyAsynStatus(status));

    // Log the error message and try to propagate it
    asynPrint(lowLevelPortUser_, ASYN_TRACE_ERROR, "%s: %s", functionName,
              message);
    setStringParam(messageText_, message);
    return status;
}

asynStatus newPmacV3Controller::checkNumExpectedReads(int expected, int read,
                                                      const char *functionName,
                                                      const char *command,
                                                      const char *response,
                                                      int axisNo_) {
    if (expected == read) {
        return asynSuccess;
    } else {
        char message[MAXBUF_] = {0};
        snprintf(message, sizeof(message),
                 "Could not interpret response %s for command %s (axis %d). "
                 "This is a bug.",
                 response, command, axisNo_);
        asynPrint(lowLevelPortUser_, ASYN_TRACE_ERROR, "%s: %s", functionName,
                  message);
        setStringParam(messageText_, message);
        setIntegerParam(motorStatusCommsError_, 1);
        return asynError;
    }
}

/*************************************************************************************/
/** The following functions are C-wrappers, and can be called directly from
 * iocsh */

extern "C" {

/*
C wrapper for the Controller constructor.
*/
asynStatus newPmacV3CreateController(const char *portName,
                                     const char *lowLevelPortName, int numAxes,
                                     double movingPollPeriod,
                                     double idlePollPeriod) {
    /*
    We create a new instance of CreateController, using the "new" keyword to
    allocate it on the heap while avoiding RAII. TBD: Where is the pointer to
    the controller stored?
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorController.cpp
    https://github.com/epics-modules/asyn/blob/master/asyn/asynPortDriver/asynPortDriver.cpp

    Setting the pointer to nullptr / NULL immediately after construction is
    simply done to avoid compiler warnings, see page 7 of this document:
    https://subversion.xray.aps.anl.gov/synApps/measComp/trunk/documentation/measCompTutorial.pdf
    */
    newPmacV3Controller *pController = new newPmacV3Controller(
        portName, lowLevelPortName, numAxes, movingPollPeriod, idlePollPeriod);
    pController = NULL;

    return asynSuccess;
}

/*
C wrapper for the Axis constructor.
See Axis::Axis.
*/
asynStatus newPmacV3CreateAxis(const char *port, int axis) {
    newPmacV3Axis *pAxis;

    static const char *functionName = "newPmacV3CreateAxis";

    /*
    findAsynPortDriver is a asyn library FFI function which uses the C ABI.
    Therefore it returns a void pointer instead of e.g. a pointer to a
    superclass of the controller such as asynPortDriver. Type-safe upcasting via
    dynamic_cast is therefore not possible directly. However, we do know that
    the void pointer is either a pointer to asynPortDriver (if a driver with the
    specified name exists) or a nullptr. Therefore, we first do a nullptr check,
    then a cast to asynPortDriver and lastly a (typesafe) dynamic_upcast to
    Controller
    https://stackoverflow.com/questions/70906749/is-there-a-safe-way-to-cast-void-to-class-pointer-in-c
     */
    void *ptr = findAsynPortDriver(port);
    if (ptr == nullptr) {
        /*
        We can't use asynPrint here since this macro would require us
        to get a lowLevelPortUser_ from a pointer to an asynPortDriver.
        However, the given pointer is a nullptr and therefore doesn't
        have a lowLevelPortUser_! printf is an EPICS alternative which
        works w/o that, but doesn't offer the comfort provided
        by the asynTrace-facility
        */
        printf("%s:%s: Error port %s not found\n", driverName, functionName,
               port);
        return asynError;
    }
    // Unsafe cast of the pointer to an asynPortDriver
    asynPortDriver *apd = (asynPortDriver *)(ptr);

    // Safe downcast
    newPmacV3Controller *pC = dynamic_cast<newPmacV3Controller *>(apd);
    if (pC == nullptr) {
        printf("%s: controller on port %s is not a Controller\n", functionName,
               port);
        return asynError;
    }

    // Prevent manipulation of the controller from other threads while we create
    // the new axis.
    pC->lock();

    /*
    We create a new instance of Axis, using the "new" keyword to allocate it
    on the heap while avoiding RAII. In the constructor, a pointer to the new
    object is stored in the controller object "pC". Therefore, the axis instance
    can still be reached later by quering "pC".

    Setting the pointer to nullptr / NULL immediately after construction is
    simply done to avoid compiler warnings, see page 7 of this document:
    https://subversion.xray.aps.anl.gov/synApps/measComp/trunk/documentation/measCompTutorial.pdf
    */
    pAxis = new newPmacV3Axis(pC, axis);
    pAxis = nullptr;

    // Allow manipulation of the controller again
    pC->unlock();
    return asynSuccess;
}

/*
This is boilerplate code which is used to make the FFI functions
CreateController and CreateAxis "known" to the IOC shell (iocsh).
TBD: If the code is compiled for running on vxWorks, this registration is
apparently not necessary?
*/

#ifdef vxWorks
#else

/*
Define name and type of the arguments for the CreateController function
in the iocsh. This is done by creating structs with the argument names and types
and then providing "factory" functions (configCreateControllerCallFunc).
These factory functions are used to register the constructors during
compilation.
*/
static const iocshArg CreateControllerArg0 = {"Controller port name",
                                              iocshArgString};
static const iocshArg CreateControllerArg1 = {"Low level port name",
                                              iocshArgString};
static const iocshArg CreateControllerArg2 = {"Number of axes", iocshArgInt};
static const iocshArg CreateControllerArg3 = {"Moving poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg4 = {"Idle poll rate (s)",
                                              iocshArgDouble};
static const iocshArg *const CreateControllerArgs[] = {
    &CreateControllerArg0, &CreateControllerArg1, &CreateControllerArg2,
    &CreateControllerArg3, &CreateControllerArg4};
static const iocshFuncDef configNewPmacV3CreateController = {
    "newPmacV3CreateController", 5, CreateControllerArgs};
static void configNewPmacV3CreateControllerCallFunc(const iocshArgBuf *args) {
    newPmacV3CreateController(args[0].sval, args[1].sval, args[2].ival,
                              args[3].dval, args[4].dval);
}

/*
Same procedure as for the CreateController function, but for the axis
itself.
*/
static const iocshArg CreateAxisArg0 = {"Controller port name", iocshArgString};
static const iocshArg CreateAxisArg1 = {"Axis number", iocshArgInt};
static const iocshArg *const CreateAxisArgs[] = {&CreateAxisArg0,
                                                 &CreateAxisArg1};
static const iocshFuncDef configNewPmacV3CreateAxis = {"newPmacV3CreateAxis", 2,
                                                       CreateAxisArgs};
static void configNewPmacV3CreateAxisCallFunc(const iocshArgBuf *args) {
    newPmacV3CreateAxis(args[0].sval, args[1].ival);
}

// This function is made known to EPICS in sinq.dbd and is called by EPICS
// in order to register both functions in the IOC shell
// TBD: Does this happen during compilation?
static void newPmacV3ControllerRegister(void) {
    iocshRegister(&configNewPmacV3CreateController,
                  configNewPmacV3CreateControllerCallFunc);
    iocshRegister(&configNewPmacV3CreateAxis,
                  configNewPmacV3CreateAxisCallFunc);
}
epicsExportRegistrar(newPmacV3ControllerRegister);

#endif

} // extern "C"
