/********************************************
 *  pmacV3Controller.h
 *
 * PMAC V3 controller driver based on the asynMotorController class
 *
 * Stefan Mathis, September 2024
 ********************************************/

#ifndef pmacV3Controller_H
#define pmacV3Controller_H
#include "asynMotorAxis.h"
#include "asynMotorController.h"
#include "newPmacV3Axis.h"

#define IncrementalEncoder "Incremental encoder"
#define AbsoluteEncoder "Absolute encoder"

class newPmacV3Controller : public asynMotorController {

  public:
    newPmacV3Controller(const char *portName, const char *lowLevelPortName,
                        int numAxes, double movingPollPeriod,
                        double idlePollPeriod, const int &extraParams = 2);

    virtual ~newPmacV3Controller();

    /* Overloaded methods methods */
    newPmacV3Axis *getAxis(asynUser *pasynUser);
    newPmacV3Axis *getAxis(int axisNo);

    // overloaded because we want to enable/disable the motor
    asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);

    // overloaded because we want to read the axis state
    asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);

    // // Overloaded to read configuration details from the motor records
    // asynStatus drvUserCreate(asynUser *pasynUser, const char *drvInfo,
    //                          const char **pptypeName, size_t *psize);

  protected:
    asynUser *lowLevelPortUser_;

    asynStatus writeRead(int axisNo, const char *command, char *response,
                         bool expect_response);

    // Create a descriptive string out of an asynStatus which can be used for
    // logging or communicating with the user
    const char *stringifyAsynStatus(asynStatus status);
    asynStatus paramLibAccessFailed(asynStatus status, const char *functionName,
                                    const char *parameter);
    asynStatus checkNumExpectedReads(int expected, int read,
                                     const char *functionName,
                                     const char *command, const char *response,
                                     int axisNo_);
    newPmacV3Axis *castToAxis(asynMotorAxis *asynAxis);

  private:
    // Set the maximum buffer size. This is an empirical value which must be
    // large enough to avoid overflows for all commands to the device /
    // responses from it.
    static const uint32_t MAXBUF_ = 200;

    /*
    When trying to communicate with the device, the underlying asynOctetSyncIO
    interface waits for a response until this time (in seconds) has passed,
    then it declares a timeout. This variable has to be specified in the
    .cpp-file.
    */
    static const double TIMEOUT_;

    // Indices of additional PVs
    int messageText_;
    int enableMotor_;
    int motorEnabled_;
    int rereadEncoderPosition_;
    int rereadEncoderPositionRBV_;
    int readConfig_;
    int encoderType_;
    int motorPositionRBV_;

    /*
    If the time between two sent messages is too short, the MCU communication
    module might "lose" an answer. To prevent this, a small delay is introduced
    in EPICS after each message exchange. Unit is microseconds.
    */
    int afterMessageSleep_;

    friend class newPmacV3Axis;
};

#endif /* pmacV3Controller_H */
